@@ -9,13 +9,9 @@
 # the mode switching program with the matching parameter
 # file from /usr/share/usb_modeswitch
 #
-# Part of usb-modeswitch-1.2.0 package
+# Part of usb-modeswitch-1.2.1 package
 # (C) Josua Dietze 2009, 2010, 2011
 
-
-# Setting of these switches is done in the global config
-# file (/etc/usb_modeswitch.conf) if available
-
 set arg0 [lindex $argv 0]
 if [regexp {\.tcl$} $arg0] {
 	if [file exists $arg0] {
@@ -25,39 +21,66 @@ if [regexp {\.tcl$} $arg0] {
 	}
 }
 
-set logging 0
-set noswitching 0
+# Setting of these switches is done in the global config
+# file (/etc/usb_modeswitch.conf) if available
 
-set env(PATH) "/bin:/usr/bin"
+set flags(logging) 0
+set flags(noswitching) 0
+
+set env(PATH) "/bin:/sbin:/usr/bin:/usr/sbin"
 
 # Execution starts at file bottom
 
 proc {Main} {argv argc} {
 
-global scsi usb config match wc device logging noswitching settings
+global scsi usb config match device flags settings
 
 set loginit [ParseGlobalConfig]
 
 # The facility to add a symbolic link pointing to the
 # ttyUSB port which provides interrupt transfer, i.e.
-# the port to connect through; returns a symlink name
-# for udev and exits
-# This is run once for every known device interface by
+# the port to connect through.
+# Will check for interrupt endpoint in ttyUSB port (lowest if
+# there is more than one); if found, return "gsmmodem[n]" name
+# to udev for symlink creation
+
+# This is run once for every port of LISTED devices by
 # an udev rule
 
 if {[lindex $argv 0] == "--symlink-name"} {
-	if $logging {
-		set device [clock clicks]
-	}
 	puts -nonewline [SymLinkName [lindex $argv 1]]
 	SafeExit
 }
 
+set argList [split [lindex $argv 1] /]
+
+if [string length [lindex $argList 1]] {
+	set device [lindex $argList 1]
+} else {
+	set device "noname"
+}
+
+Log "Raw args from udev: [lindex $argv 1]\n\n$loginit"
+
+if {$device == "noname"} {
+	Log "No data from udev. Exiting"
+	SafeExit
+}
+
+if {[lindex $argv 0] != "--switch-mode"} {
+	Log "No command given. Exiting"
+	SafeExit
+}
+
+if {![regexp /lib/udev/usb_modeswitch [lindex $argv 2]]} {
+	Log "Dispatcher was not run from call script. Exiting"
+	SafeExit
+}
+
 set settings(dbdir)	/usr/share/usb_modeswitch
 set settings(dbdir_etc)	/etc/usb_modeswitch.d
 
 if {![file exists $settings(dbdir)] && ![file exists $settings(dbdir_etc)]} {
-	set device "noname"
 	Log "Error: no config database found in /usr/share or /etc. Exiting"
 	SafeExit
 }
@@ -66,23 +89,6 @@ set bindir /usr/sbin
 set devList1 {}
 set devList2 {}
 
-# argv contains the values provided from the udev rule
-# separated by "/"
-
-set argList [split [lindex $argv 0] /]
-
-if [string length [lindex $argList 1]] {
-	set device [lindex $argList 1]
-} else {
-	set device "noname"
-}
-
-Log "Raw args from udev: $argv\n\n$loginit"
-
-if {$device == "noname"} {
-	Log "No data from udev. Exiting"
-	SafeExit
-}
 
 # arg 0: the bus id for the device (udev: %b)
 # arg 1: the "kernel name" for the device (udev: %k)
@@ -98,8 +104,6 @@ if {[string length [lindex $argList 0]] == 0} {
 		Log "No device number values given from udev! Exiting"
 		SafeExit
 	} else {
-		Log "Bus ID for device not given by udev."
-		Log " Trying to determine it from kernel name ([lindex $argList 1]) ..."
 		if {![regexp {(.*?):} [lindex $argList 1] d dev_top]} {
 			Log "Could not determine top device dir from udev values! Exiting"
 			SafeExit
@@ -110,12 +114,13 @@ if {[string length [lindex $argList 0]] == 0} {
 	regexp {(.*?):} $dev_top d dev_top
 }
 
-
 set devdir /sys/bus/usb/devices/$dev_top
 if {![file isdirectory $devdir]} {
-	Log "Top sysfs directory not found ($devdir)! Exiting"
+	Log "Top device directory not found ($devdir)! Exiting"
 	SafeExit
 }
+Log "Using top device dir $devdir"
+set ifdir "[file tail $devdir]:1.0"
 
 
 # Mapping of the short string identifiers (in the config
@@ -132,21 +137,20 @@ set match(uSe) usb(serial)
 
 
 # Now reading the USB attributes
-
-ReadUSBAttrs $devdir
-
-if {[string length "$usb(idVendor)$usb(idProduct)"] < 8} {
-	Log "USB IDs not found in sysfs tree. Exiting"
+if {![ReadUSBAttrs $devdir]} {
+	Log "USB attributes not found in sysfs tree. Exiting"
 	SafeExit
 }
 
-Log "----------------\nUSB values from sysfs:"
-foreach attr {manufacturer product serial} {
-	Log "  $attr\t$usb($attr)"
+if $flags(logging) {
+	Log "----------------\nUSB values from sysfs:"
+	foreach attr {manufacturer product serial} {
+		Log "  $attr\t$usb($attr)"
+	}
+	Log "----------------"
 }
-Log "----------------"
 
-if $noswitching {
+if $flags(noswitching) {
 	Log "\nSwitching globally disabled. Exiting\n"
 	catch {exec logger -p syslog.notice "usb_modeswitch: switching disabled, no action for $usb(idVendor):$usb(idProduct)" 2>/dev/null}
 	SafeExit
@@ -176,107 +180,25 @@ if {[llength $configList] > 1} {
 		set scsiNeeded 1
 	}
 }
-if {!$scsiNeeded} {
-	Log "SCSI attributes not needed, moving on"
-}
-
-
-# Getting the SCSI values via libusb results in a detached
-# usb-storage driver. Not good for devices that want to be
-# left alone. Fortunately, the sysfs tree provides the values
-# too without need for direct access
-
-# First we wait until the SCSI data is ready - or timeout.
-# Timeout means: no storage driver was bound to the device.
-# We run 20 times max, every half second (max. 10 seconds
-# total)
-
-# We also check if the device itself changes, probably
-# because it was switched by the kernel (or even unplugged).
-# Then we do simply nothing and exit quietly ...
-
-set counter 0
-while {$scsiNeeded && $counter < 20} {
-	after 500
-	incr counter
-	Log "waiting for storage tree in sysfs"
-
-	set sysdir $devdir/[lindex $argList 1]
-
-	if {![file isdirectory $sysdir]} {
-		# Device is gone. Unplugged? Switched by kernel?
-		Log "sysfs device tree is gone; exiting"
-		SafeExit
-	}
-	set rc [open $devdir/product r]
-	set newproduct [read -nonewline $rc]
-	close $rc
-	if {![string match $newproduct $usb(product)]} {
-		# Device has just changed. Switched by someone else?
-		Log "device has changed; exiting"
-		SafeExit
-	}
-
-	# Searching the storage/SCSI tree; might take a while
-	if {[set dirList [glob -nocomplain $sysdir/host*]] != ""} {
-		set sysdir [lindex $dirList 0]
-		if {[set dirList [glob -nocomplain $sysdir/target*]] != ""} {
-			set sysdir [lindex $dirList 0]
-			regexp {.*target(.*)} $sysdir d subdir
-			if {[set dirList [glob -nocomplain $sysdir/$subdir*]] != ""} {
-				set sysdir [lindex $dirList 0]
-				if [file exists $sysdir/vendor] {
-					# Finally SCSI structure is ready, get the values
-					ReadSCSIAttrs $sysdir
-					Log "SCSI values read"
-					break
-				}
-			}
-		}
-	}
-}
 if $scsiNeeded {
-	if {$counter == 20 && [string length $scsi(vendor)] == 0} {
-		Log "SCSI tree not found; you may want to check if this path/file exists:"
-		Log "$sysdir/vendor\n"
-	} else {
+	if [ReadSCSIAttrs $devdir:1.0] {
 		Log "----------------\nSCSI values from sysfs:"
 		foreach attr {vendor model rev} {
 			Log " $attr\t$scsi($attr)"
 		}
 		Log "----------------"
+	} else {
+		Log "Could not get SCSI attributes, exclude devices with SCSI match"
 	}
-	Log "Waiting 2 secs. after SCSI device was added"
-	after 2000
 } else {
-	after 500
+	Log "SCSI attributes not needed, moving on"
 }
 
-# If SCSI tree in sysfs was not identified, try and get the values
-# from a (nonswitching) call of usb_modeswitch; this detaches the
-# storage driver, so it's just the last resort
+# General wait - this is important
+after 500
 
-if {$scsiNeeded && $scsi(vendor)==""} {
-	set testSCSI [exec $bindir/usb_modeswitch -v 0x$usb(idVendor) -p 0x$usb(idProduct) 2>/dev/null]
-	regexp {  Vendor String: (.*?)\n} $testSCSI d scsi(vendor)
-	regexp {   Model String: (.*?)\n} $testSCSI d scsi(model)
-	regexp {Revision String: (.*?)\n} $testSCSI d scsi(rev)
-	Log "SCSI values from usb_modeswitch:"
-	foreach attr {vendor model rev} {
-		Log " $attr\t$scsi($attr)"
-	}
-}
-
-# If we don't have the SCSI values by now, we just
-# leave the variables empty; they won't match anything
-
-# Time to check for a matching config file.
-# Matching itself is done by MatchDevice
-#
-# The configuration file names are sorted reverse so that
-# the ones with matching additions are tried first; the
-# common configs without match attributes are used at the
-# end and provide a fallback
+# Now check for a matching config file. Matching is done
+# by MatchDevice
 
 set report {}
 foreach configuration $configList {
@@ -286,102 +208,100 @@ foreach configuration $configList {
 
 	Log "checking config: $configuration"
 	if [MatchDevice $configuration] {
+		Log "! matched. Reading config data"
+		if [string length $usb(busnum)] {
+			set busParam "-b [string trimleft $usb(busnum) 0]"
+			set devParam "-g [string trimleft $usb(devnum) 0]"
+		} else {
+			set busParam ""
+			set devParam ""
+		}
 		set configBuffer [ConfigGet conffile $configuration]
 		ParseDeviceConfig $configBuffer
 		set devList1 [ListSerialDevs]
 		if {$config(waitBefore) == ""} {
-			Log "! matched, now switching"
 		} else {
-			Log "! matched, waiting time set to $config(waitBefore) seconds"
+			Log " waiting time set to $config(waitBefore) seconds"
 			append config(waitBefore) "000"
 			after $config(waitBefore)
 			Log " waiting is over, switching starts now"
 		}
 
 		# Now we are actually switching
-		if $logging {
-			Log " (running command: $bindir/usb_modeswitch -I -W -D $configParam -f \$configBuffer)"
-			set report [exec $bindir/usb_modeswitch -I -W -D $configParam -f "$configBuffer" 2>@ stdout]
+		if $flags(logging) {
+			Log "Command to be run:\nusb_modeswitch -I -W -D -s 20 $configParam $busParam $devParam -v $usb(idVendor) -p $usb(idProduct) -f \$configBuffer"
+			set report [exec usb_modeswitch -I -W -D -s 20 $configParam $busParam $devParam -v $usb(idVendor) -p $usb(idProduct) -f "$configBuffer" 2>@ stdout]
+			Log "\nVerbose debug output of usb_modeswitch and libusb follows"
+			Log "(Note that some USB errors are to be expected in the process)"
+			Log "--------------------------------"
+			Log $report
+			Log "--------------------------------"
+			Log "(end of usb_modeswitch output)\n"
 		} else {
-			exec $bindir/usb_modeswitch -I -Q -D $configParam -f "$configBuffer" 2>/dev/null]
+			set report [exec usb_modeswitch -I -Q -D -s 20 $configParam $busParam $devParam -v $usb(idVendor) -p $usb(idProduct) -f "$configBuffer" 2>@ stdout]
 		}
-		Log "\nVerbose debug output of usb_modeswitch and libusb follows"
-		Log "(Note that some USB errors are expected in the process)"
-		Log "--------------------------------"
-		Log $report
-		Log "--------------------------------"
-		Log "(end of usb_modeswitch output)\n"
 		break
 	} else {
 		Log "* no match, not switching with this config"
 	}
 }
 
-# We're finished with switching; success checking
-# was done by usb_modeswitch and logged via syslog.
-#
+# Switching is complete; success checking was either
+# done by usb_modeswitch and logged via syslog OR bus/dev
+# parameter were used; then we do check for success HERE
+
+if [regexp {ok:busdev} $report] {
+	if [CheckSuccess $devdir] {
+		Log "Mode switching was successful, found $usb(idVendor):$usb(idProduct) ($usb(manufacturer): $usb(product))"
+		set syslog_text "usb_modeswitch: switched to $usb(idVendor):$usb(idProduct) on [format %03d $usb(busnum)]/[format %03d $usb(devnum)]"
+		catch {exec logger -p syslog.notice $syslog_text}
+	} else {
+		Log "Mode switching may have failed. Exiting"
+		SafeExit
+	}
+} else {
+	if {![file isdirectory $devdir]} {
+		Log "Device directory in sysfs is gone! Something went wrong, aborting"
+		SafeExit
+	}
+	# Give the device another second if it's not fully back yet
+	if {![file exists $devdir/idProduct]} {
+		after 1000
+	}
+	ReadUSBAttrs $devdir $ifdir
+}
+
 # If switching was OK we now check for drivers by
 # simply recounting serial devices under /dev
 
-if {![file isdirectory $devdir]} {
-	Log "Device directory in sysfs is gone! Something went wrong, aborting"
-	SafeExit
-}
-# Give the device another second if it's not fully back yet
-if {![file exists $devdir/idProduct]} {
-	after 1000
-}
-
-set ifdir "[file tail $devdir]:1.0"
-ReadUSBAttrs $devdir $ifdir
-
 if {$usb($ifdir/bInterfaceClass) != "" && [regexp {ok:} $report]} {
 	if {$usb($ifdir/bInterfaceClass) == "02"} {
-		set report "ok:"
+		set config(driverModule) ""
 		Log " Found CDC ACM device, skip driver checking"
 	}
 }
 
-# If target ID given, driver shall be loaded
-if [regexp -nocase {ok:[0-9a-f]{4}:[0-9a-f]{4}|ok:no_data} $report] {
+# If module is set (it is by default), driver shall be loaded.
+# If not, then NoDriverLoading is active
 
+if {$config(driverModule) != ""} {
 	if {[string length "$usb(idVendor)$usb(idProduct)"] < 8} {
-		regexp {ok:(\w{4}):(\w{4})} $report d usb(idVendor) usb(idProduct)
-	}
-	set t "$usb(idVendor)$usb(idProduct)"
-	if {[string length $t] != 8 || [string trim $t 0] == ""} {
-		if {$report == "ok:no_data"} {
-			Log "Libusb1 bug prevented device searching, and device ID not found afterwards."
-		}
-		Log "No vendor/product ID found or given, can't continue. Aborting"
-		SafeExit
-	}
-
-	# For general driver loading; TODO: add respective device names.
-	# Presently only useful for HSO devices (which are recounted now)
-	if {$config(driverModule) == ""} {
-		set config(driverModule) "option"
-		set config(driverIDPath) "/sys/bus/usb-serial/drivers/option1"
-	} else {
-		if {$config(driverIDPath) == ""} {
-			set config(driverIDPath) "/sys/bus/usb/drivers/$config(driverModule)"
+		if {![regexp {ok:(\w{4}):(\w{4})} $report d usb(idVendor) usb(idProduct)]} {
+			Log "No target vendor/product ID found or given, can't continue. Aborting"
+			SafeExit
 		}
 	}
-	Log "Driver module is \"$config(driverModule)\", ID path is $config(driverIDPath)\n"
-
 	# wait for any drivers to bind automatically
 	after 1000
-
-	Log "Now checking for newly created serial devices ..."
+	Log "Now checking for newly created ports ..."
 	set devList2 [ListSerialDevs]
-
 	if {[llength $devList1] >= [llength $devList2]} {
-		Log " no new serial devices found"
+		Log " no new ports found"
 		AddToList link_list $usb(idVendor):$usb(idProduct)
 
 		# If device is known, the sh wrapper will take care, else:
 		if {[InBindList $usb(idVendor):$usb(idProduct)] == 0} {
-			Log "Device not in bind_list"
+			Log "Device is not in \"bind_list\" yet"
 
 			# Load driver
 			CheckDriverBind $usb(idVendor) $usb(idProduct)
@@ -394,14 +314,14 @@ if [regexp -nocase {ok:[0-9a-f]{4}:[0-9a-f]{4}|ok:no_data} $report] {
 				incr counter
 			}
 			if {$counter == 14} {
-				Log " still no new serial devices found"
+				Log " after binding, no new ports found"
 			} else {
-				Log " driver successfully bound"
+				Log " driver was bound to the device"
 				AddToList bind_list $usb(idVendor):$usb(idProduct)
 			}
 		}
 	} else {
-		Log " new serial devices found, driver has bound"
+		Log " new ports found, device is known to driver"
 		if {[llength [lsearch -glob -all $devList2 *ttyUSB*]] > [llength [lsearch -glob -all $devList1 *ttyUSB*]]} {
 			AddToList link_list $usb(idVendor):$usb(idProduct)
 		}
@@ -420,7 +340,7 @@ if [regexp {ok:$} $report] {
 # reset (e.g. from usb-storage) which would possibly switch back
 # a mode-switching device to initial mode
 if [regexp {ok:} $report] {
-	Log "Checking for AVOID_RESET_QUIRK attribute"
+	Log "Checking for AVOID_RESET_QUIRK kernel attribute"
 	if [file exists $devdir/avoid_reset_quirk] {
 		if [catch {exec echo "1" >$devdir/avoid_reset_quirk 2>/dev/null} err] {
 			Log " Error setting the attribute: $err"
@@ -428,7 +348,7 @@ if [regexp {ok:} $report] {
 			Log " AVOID_RESET_QUIRK activated"
 		}
 	} else {
-		Log " AVOID_RESET_QUIRK not present"
+		Log " not present in this kernel"
 	}
 }
 
@@ -439,14 +359,47 @@ SafeExit
 # end of proc {Main}
 
 
-proc {ReadSCSIAttrs} {dir} {
+proc {ReadSCSIAttrs} {topdir} {
 
 global scsi
-Log "SCSI dir exists: $dir"
+set counter 0
+set sysdir $topdir
+Log "Checking storage tree in sysfs ..."
+while {$counter < 20} {
+	Log " loop $counter/20"
+	if {![file isdirectory $sysdir]} {
+		# Device is gone. Unplugged? Switched by kernel?
+		Log " sysfs device tree is gone; abort SCSI value check"
+		return 0
+	}
+	# Searching the storage/SCSI tree; might take a while
+	if {[set dirList [glob -nocomplain $topdir/host*]] != ""} {
+		set sysdir [lindex $dirList 0]
+		if {[set dirList [glob -nocomplain $sysdir/target*]] != ""} {
+			set sysdir [lindex $dirList 0]
+			regexp {.*target(.*)} $sysdir d subdir
+			if {[set dirList [glob -nocomplain $sysdir/$subdir*]] != ""} {
+				set sysdir [lindex $dirList 0]
+				if [file exists $sysdir/vendor] {
+					Log " Storage tree is ready"
+					break
+				}
+			}
+		}
+	}
+	after 500
+	incr counter
+}
+if {$counter == 20} {
+	Log "SCSI tree not found; you may want to check if this path/file exists:"
+	Log "$sysdir/vendor\n"
+	return 0
+}
 
+Log "Reading SCSI values ..."
 foreach attr {vendor model rev} {
-	if [file exists $dir/$attr] {
-		set rc [open $dir/$attr r]
+	if [file exists $sysdir/$attr] {
+		set rc [open $sysdir/$attr r]
 		set scsi($attr) [read -nonewline $rc]
 		close $rc
 	} else {
@@ -454,6 +407,7 @@ foreach attr {vendor model rev} {
 		Log "Warning: SCSI attribute \"$attr\" not found."
 	}
 }
+return 1
 
 }
 # end of proc {ReadSCSIAttrs}
@@ -463,20 +417,27 @@ proc {ReadUSBAttrs} {dir args} {
 
 global usb
 
-set attrList {idVendor idProduct manufacturer product serial bNumConfigurations}
+set attrList {idVendor idProduct bConfigurationValue manufacturer product serial devnum busnum bNumConfigurations}
+set mandatoryList {idVendor idProduct bNumConfigurations}
+set result 1
 if {$args != ""} {
 	lappend attrList "$args/bInterfaceClass"
+	lappend mandatoryList "$args/bInterfaceClass"
 }
 foreach attr $attrList {
 	if [file exists $dir/$attr] {
 		set rc [open $dir/$attr r]
-		set usb($attr) [read -nonewline $rc]
+		set usb($attr) [string trim [read -nonewline $rc]]
 		close $rc
 	} else {
 		set usb($attr) ""
+		if {[lsearch $mandatoryList $attr] > -1} {
+			set result 0
+		}
 		Log "Warning: USB attribute \"$attr\" not found."
 	}
 }
+return $result
 
 }
 # end of proc {ReadUSBAttrs}
@@ -514,8 +475,7 @@ return 1
 
 proc {ParseGlobalConfig} {} {
 
-global logging noswitching
-
+global flags
 set configFile ""
 set places [list /etc/usb_modeswitch.conf /etc/sysconfig/usb_modeswitch /etc/default/usb_modeswitch]
 foreach cfg $places {
@@ -524,7 +484,6 @@ foreach cfg $places {
 		break
 	}
 }
-
 if {$configFile == ""} {return}
 
 set rc [open $configFile r]
@@ -532,12 +491,12 @@ while {![eof $rc]} {
 	gets $rc line
 	if [regexp {DisableSwitching\s*=\s*([^\s]+)} $line d val] {
 		if [regexp -nocase {1|yes|true} $val] {
-			set noswitching 1
+			set flags(noswitching) 1
 		}
 	}
 	if [regexp {EnableLogging\s*=\s*([^\s]+)} $line d val] {
 		if [regexp -nocase {1|yes|true} $val] {
-			set logging 1
+			set flags(logging) 1
 		}
 	}
 
@@ -554,17 +513,59 @@ global config
 set config(driverModule) ""
 set config(driverIDPath) ""
 set config(waitBefore) ""
-set lineList [split $configContent \n]
-foreach line $lineList {
-	if [regexp {^DriverModule[[:blank:]]*=[[:blank:]]*"?(\w+)"?} [string trim $line] d config(driverModule)] {
-		Log "config: DriverModule set to $config(driverModule)"
-	}
-	if [regexp {^DriverIDPath[[:blank:]]*=[[:blank:]]*?"?([/\-\w]+)"?} [string trim $line] d config(driverIDPath)] {
-		Log "config: DriverIDPath set to $config(driverIDPath)"
-	}
-	if [regexp {^WaitBefore[[:blank:]]*=[[:blank:]]*?([0-9]+)} [string trim $line] d config(waitBefore)] {
-		Log "config: WaitBefore set to $config(waitBefore)"
+set config(targetVendor) ""
+set config(targetProduct) ""
+set config(targetClass) ""
+set config(Configuration) ""
+set config(checkSuccess) 20
+set loadDriver 1
+
+if [regexp -line {^[[:blank:]]*TargetVendor[[:blank:]]*=[[:blank:]]*"?0x(\w+)"?.*?$} $configContent d config(targetVendor)] {
+	Log "config: TargetVendor set to $config(targetVendor)"
+}
+if [regexp -line {^[[:blank:]]*TargetProduct[[:blank:]]*=[[:blank:]]*"?0x(\w+)"?.*?$} $configContent d config(targetProduct)] {
+	Log "config: TargetProduct set to $config(targetProduct)"
+}
+if [regexp -line {^[[:blank:]]*TargetProductList[[:blank:]]*=[[:blank:]]*"?([0-9a-fA-F,]+)"?.*?$} $configContent d config(targetProduct)] {
+	Log "config: TargetProductList set to $config(targetProduct)"
+}
+if [regexp -line {^[[:blank:]]*TargetClass[[:blank:]]*=[[:blank:]]*"?(0x\w+)"?.*?$} $configContent d config(targetClass)] {
+	Log "config: TargetClass set to $config(targetClass)"
+}
+if [regexp -line {^[[:blank:]]*Configuration[[:blank:]]*=[[:blank:]]*([0-9]+).*?$} $configContent d config(Configuration)] {
+	Log "config: Configuration (target) set to $config(Configuration)"
+}
+if [regexp -line {^[[:blank:]]*DriverModule[[:blank:]]*=[[:blank:]]*"?(\w+)"?.*?$} $configContent d config(driverModule)] {
+	Log "config: DriverModule set to $config(driverModule)"
+}
+if [regexp -line {^[[:blank:]]*DriverIDPath[[:blank:]]*=[[:blank:]]*?"?([/\-\w]+)"?.*?$} $configContent d config(driverIDPath)] {
+	Log "config: DriverIDPath set to $config(driverIDPath)"
+}
+if [regexp -line {^[[:blank:]]*CheckSuccess[[:blank:]]*=[[:blank:]]*?([0-9]+).*?$} $configContent d config(checkSuccess)] {
+	Log "config: CheckSuccess set to $config(checkSuccess)"
+}
+if [regexp -line {^[[:blank:]]*WaitBefore[[:blank:]]*=[[:blank:]]*?([0-9]+).*?$} $configContent d config(waitBefore)] {
+	Log "config: WaitBefore set to $config(waitBefore)"
+}
+if [regexp -line {^[[:blank:]]*NoDriverLoading[[:blank:]]*=[[:blank:]]*?(1|yes|true).*?$} $configContent] {
+	set loadDriver 0
+	Log "config: NoDriverLoading is set to active"
+}
+
+# For general driver loading; TODO: add respective device names.
+# Presently only useful for HSO devices (which are recounted now)
+if $loadDriver {
+	if {$config(driverModule) == ""} {
+		set config(driverModule) "option"
+		set config(driverIDPath) "/sys/bus/usb-serial/drivers/option1"
+	} else {
+		if {$config(driverIDPath) == ""} {
+			set config(driverIDPath) "/sys/bus/usb/drivers/$config(driverModule)"
+		}
 	}
+	Log "Driver module is \"$config(driverModule)\", ID path is $config(driverIDPath)\n"
+} else {
+	Log "Driver will not be handled by usb_modeswitch"
 }
 set config(waitBefore) [string trimleft $config(waitBefore) 0]
 
@@ -601,12 +602,8 @@ switch $command {
 	conffile {
 		if [regexp {^pack/} $config] {
 			set config [regsub {pack/} $config {}]
-#			set settings(tmpConfig) /var/lib/usb_modeswitch/current_cfg
 			Log "Extracting config $config from collection $settings(dbdir)/configPack.tar.gz"
 			set configContent [exec tar -xzOf $settings(dbdir)/configPack.tar.gz $config 2>/dev/null]
-#			set wc [open $settings(tmpConfig) w]
-#			puts -nonewline $wc [exec tar -xzOf $settings(dbdir)/configPack.tar.gz $config 2>/dev/null]
-#			close $wc
 		} else {
 			if [regexp [list $settings(dbdir_etc)] $config] {
 				Log "Using config file from override folder $settings(dbdir_etc)"
@@ -628,22 +625,22 @@ switch $command {
 
 proc {Log} {msg} {
 
-global wc logging device
-if {$logging == 0} {return}
+global flags device
+if {$flags(logging) == 0} {return}
 
-if {![info exists wc]} {
-	if [catch {set wc [open /var/log/usb_modeswitch_$device w]} err] {
-		if [catch {set wc [open /dev/console w]} err] {
-			set wc "error"
+if {![info exists flags(wc)]} {
+	if [catch {set flags(wc) [open /var/log/usb_modeswitch_$device w]} err] {
+		if [catch {set flags(wc) [open /dev/console w]} err] {
+			set flags(wc) "error"
 			return
 		} else {
-			puts $wc "Error opening log file ($err), redirect to console"
+			puts $flags(wc) "Error opening log file ($err), redirect to console"
 		}
 	}
-	puts $wc "\n\nUSB_ModeSwitch log from [clock format [clock seconds]]\n"
+	puts $flags(wc) "\n\nUSB_ModeSwitch log from [clock format [clock seconds]]\n"
 }
-if {$wc == "error"} {return}
-puts $wc $msg
+if {$flags(wc) == "error"} {return}
+puts $flags(wc) $msg
 
 }
 # end of proc {Log}
@@ -652,9 +649,9 @@ puts $wc $msg
 # Closing the log file if open and exit
 proc {SafeExit} {} {
 
-global wc
-if [info exists wc] {
-	catch {close $wc}
+global flags
+if [info exists flags(wc)] {
+	catch {close $flags(wc)}
 }
 exit
 
@@ -662,29 +659,23 @@ exit
 # end of proc {SafeExit}
 
 
-# Checking for interrupt endpoint in ttyUSB port (lowest if there is
-# more than one); if found, check for unused "gsmmodem[n]" name.
-# Link for first modem will be "gsmmodem", then "gsmmodem2" and up
-#
 proc {SymLinkName} {path} {
 global device
 
-set loginit "* called with --symlink-name: params $path *\n"
-
-# Internal proc, used only here
 proc {hasInterrupt} {ifDir} {
 	if {[llength [glob -nocomplain $ifDir/ttyUSB*]] == 0} {
 		Log "  no ttyUSB interface - skip checking endpoints"
 		return 0
 	}
 	foreach epDir [glob -nocomplain $ifDir/ep_*] {
-		Log "  in epDir $epDir ..."
+		set e [file tail $epDir]
+		Log "  checking $e ..."
 		if [file exists $epDir/type] {
 			set rc [open $epDir/type r]
 			set type [read $rc]
 			close $rc
 			if [regexp {Interrupt} $type] {
-				Log " $epDir has interrupt transfer type"
+				Log "  $e has interrupt transfer type"
 				return 1
 			}
 		}
@@ -692,71 +683,56 @@ proc {hasInterrupt} {ifDir} {
 	return 0
 }
 
+set loginit "usb_modeswitch called with --symlink-name\n parameter: $path\n"
+
 # In case the device path is returned as /class/tty/ttyUSB,
-# we need to extract the USB device path from symlink "device"
+# get the USB device path from linked tree "device"
 set linkpath /sys$path/device
 if [file exists $linkpath] {
 	if {[file type $linkpath] == "link"} {
 		set rawpath [file readlink $linkpath]
 		set trimpath [regsub -all {\.\./} $rawpath {}]
  		if [file isdirectory /sys/$trimpath] {
+			append loginit "\n Using path $path\n"
 			set path /$trimpath
 		}
 	}
 }
 
-
 if {![regexp {ttyUSB[0-9]+} $path myPort]} {
-	Log "$loginit\nCould not find port name in path\n $path. Aborting"
+	if $flags(logging) {
+		set device [clock clicks]
+		Log "$loginit\nThis is not a ttyUSB port. Aborting"
+	}
 	return ""
 }
+
 set device $myPort
-Log "$loginit\nMy name is $myPort"
+Log "$loginit\nMy name is $myPort\n"
 
-if {![regexp {usb[0-9]+/([0-9]+-[0-9]+)/} $path d dev_top]} {
-	Log "Could not find device directory in path\n $path. Aborting"
+if {![regexp {(.*?[0-9]+)\.([0-9]+)/ttyUSB} /sys$path d ifRoot ifNum]} {
+	Log "Could not find interface in path\n $path. Aborting"
 	return ""
 }
 
-if {![regexp "\[0-9\]+\\.(\[0-9\]+)/$myPort" $path d myIf]} {
-	Log "Could not find interface number in path\n $path. Aborting"
-	return ""
-}
+set ifDir $ifRoot.$ifNum
 
-if {![regexp "$dev_top:\[0-9\]" /sys$path ifRoot]} {
-	Log "Could not find interface number in path\n $path. Aborting"
-	return ""
-}
-
-set dirList [split $path /]
-set idx [lsearch $dirList $dev_top]
-
-set devDir /sys[join [lrange $dirList 0 $idx] /]
-
-Log "My port is $myPort, my interface is $myIf
- devDir:  $devDir\n dev_top: $dev_top\nsysPath: /sys$path"
-
-regexp "$devDir/$dev_top:\[0-9\]" /sys$path ifRoot
-
-set ifDir $ifRoot.$myIf
-
-Log "Checking my endpoints in $ifDir"
+Log "Checking my endpoints ...\n in $ifDir"
 if [hasInterrupt $ifDir] {
-	Log "\n--> I am an interrupt port\n"
+	Log "\n--> I am an interrupt port"
 	set rightPort 1
 } else {
 	Log "\n--> I am not an interrupt port\n"
 	set rightPort 0
 }
 
-# Unfortunately, there are devices with more than one interrupt
-# port. The assumption so far is that the lowest of these is
-# right. Check all lower interfaces for another one (if interface)
-# is bigger than 0). If found, don't return any name.
+# There are devices with more than one interrupt interface.
+# Assume that the lowest of these is usable. Check all
+# possible lower interfaces
 
-if { $rightPort && ($myIf > 0) } {
-	Log "Looking for lower ports with interrupt endpoints"
-	for {set i 0} {$i < $myIf} {incr i} {
+if { $rightPort && ($ifNum > 0) } {
+	Log "\nLooking for lower ports with interrupt endpoints"
+	for {set i 0} {$i < $ifNum} {incr i} {
 		set ifDir $ifRoot.$i
 		Log " in ifDir $ifDir ..."
 		if [hasInterrupt $ifDir] {
@@ -766,7 +742,6 @@ if { $rightPort && ($myIf > 0) } {
 		}
 	}
 }
-
 if {$rightPort == 0} {
 	Log "Return empty name and exit"
 	return ""
@@ -779,11 +754,15 @@ set idx 2
 set symlinkName "gsmmodem"
 while {$idx < 256} {
 	if {![file exists $symlinkName]} {
+		set placeholder [open /dev/$symlinkName w]
+		close $placeholder
 		break
 	}
 	set symlinkName gsmmodem$idx
 	incr idx
 }
+if {$idx == 256} {return ""}
+
 Log "Return symlink name \"$symlinkName\" and exit"
 return $symlinkName
 
@@ -791,17 +770,12 @@ return $symlinkName
 # end of proc {SymLinkName}
 
 
-# Load and bind (serial) driver
+# Load and bind driver (default "option")
 #
 proc {CheckDriverBind} {vid pid} {
 global config
 
-set loader ""
-if [file exists /sbin/modprobe] {
-	set loader /sbin/modprobe
-} else {
-	Log " /sbin/modprobe not found"
-}
+set loader [exec which modprobe]
 
 set idfile $config(driverIDPath)/new_id
 if {![file exists $idfile]} {
@@ -809,14 +783,14 @@ if {![file exists $idfile]} {
 		Log "Can't do anymore without module loader; get \"modtools\"!"
 		return
 	}
-	Log "\nTrying to load driver \"$config(driverModule)\""
+	Log "\nTrying to load module \"$config(driverModule)\""
 	if [catch {set result [exec $loader -v $config(driverModule)]} err] {
 		Log " Running \"$loader $config(driverModule)\" gave an error:\n  $err"
 	} else {
-		Log " Driver was loaded successfully:\n$result"
+		Log " Module was loaded successfully:\n$result"
 	}
 } else {
-	Log "Driver was loaded already"
+	Log "Module is active already"
 }
 set i 0
 while {$i < 50} {
@@ -831,23 +805,24 @@ if {$i < 50} {
 	catch {exec logger -p syslog.notice "usb_modeswitch: adding device ID $vid:$pid to driver \"$config(driverModule)\"" 2>/dev/null}
 	catch {exec logger -p syslog.notice "usb_modeswitch: please report the device ID to the Linux USB developers!" 2>/dev/null}
 	if [catch {exec echo "$vid $pid ff" >$idfile} err] {
-		Log "Error adding ID to driver: $err"
+		Log " Error adding ID to driver:\n  $err"
 	} else {
 		Log " ID added to driver; check for new devices in /dev"
 	}
 } else {
-	Log " \"$idfile\" not found, can't add ID to driver;\n check if kernel version is at least 2.6.27"
+	Log " \"$idfile\" not found, check if kernel version is at least 2.6.27"
 	Log "Falling back to \"usbserial\""
 	set config(driverModule) usbserial
-	Log "\nTrying to unload driver \"$config(driverModule)\""
-	if [catch {exec $loader -r $config(driverModule)} err] {
-		Log " Running \"$loader $config(driverModule)\" gave an error:\n  $err"
-		Log "Can't unload usbserial. No more fallbacks"
+	Log "\nTrying to unload driver \"usbserial\""
+	if [catch {exec $loader -r usbserial} err] {
+		Log " Running \"$loader -r usbserial\" gave an error:\n  $err"
+		Log "No more fallbacks"
 		return
 	}
+	after 50
 	Log "\nTrying to load driver \"usbserial\" with device IDs"
 	if [catch {set result [exec $loader -v usbserial vendor=0x$vid product=0x$pid]} err] {
-		Log " Running \"$loader $config(driverModule)\" gave an error:\n  $err"
+		Log " Running \"$loader usbserial\" gave an error:\n  $err"
 	} else {
 		Log " Driver was loaded successfully:\n$result"
 	}
@@ -900,9 +875,9 @@ if [file exists $listfile] {
 }
 lappend idList $id
 set buffer [join $idList "\n"]
-if [catch {set wc [open $listfile w]}] {return}
-puts $wc $buffer
-close $wc
+if [catch {set lc [open $listfile w]}] {return}
+puts $lc $buffer
+close $lc
 
 }
 # end of proc {AddToList}
@@ -931,9 +906,9 @@ if {[llength $idList] == 0} {
 	return
 }
 set buffer [join $idList "\n"]
-if [catch {set wc [open $listfile w]}] {return}
-puts $wc $buffer
-close $wc
+if [catch {set lc [open $listfile w]}] {return}
+puts $lc $buffer
+close $lc
 
 }
 # end of proc {RemoveFromBindList}
@@ -950,6 +925,47 @@ return $devList
 }
 # end of proc {ListSerialDevs}
 
+proc {CheckSuccess} {devdir} {
+
+global config usb
+set ifdir "[file tail $devdir]:1.0"
+
+if {[string length $config(targetClass)] || [string length $config(Configuration)]} {
+	set config(targetVendor) $usb(idVendor)
+	set config(targetProduct) $usb(idProduct)
+}
+Log "Checking success of mode switch for max. $config(checkSuccess) seconds ..."
+
+for {set i 1} {$i <= $config(checkSuccess)} {incr i} {
+	after 1000
+	if {![file isdirectory $devdir]} {
+		Log " Waiting for device file system ($i sec.) ..."
+		continue
+	} else {
+		Log " Device file system is up, reading attributes ..."
+	}
+	if {![ReadUSBAttrs $devdir $ifdir]} {
+		Log " Essential attributes are missing, continue wait ..."
+		continue
+	}
+	if [string length $config(targetClass)] {
+		if {![regexp $usb($ifdir/bInterfaceClass) $config(targetClass)]} {continue}
+	}
+	if [string length $config(Configuration)] {
+		if {$usb(bConfigurationValue) != $config(Configuration} {continue}
+	}
+	if {![regexp $usb(idVendor) $config(targetVendor)]} {continue}
+	if {![regexp $usb(idProduct) $config(targetProduct)]} {continue}
+	Log " All attributes matched"
+	break
+}
+if {$i > 20} {
+	return 0
+}
+return 1
+
+}
+# end of proc {CheckSuccess}
 
 # The actual entry point
 Main $argv $argc
